gpt_template = """Translate the following clinical note into HL7 FHIR R4 Format. 
instructions: 
- Do not insert any values that are not in the note. 
- Do not infer or impute any values
- Only include information that is essential:
    - information that is in the clinical note
    - information that is mandatory for a valid FHIR resource.
Here is the note:
{note}
Return the result in an array of json string. denote the start and end of the json with a markdown codeblock:
```json 
[RESOURCE HERE]
```
"""


llama_template = """[INST] <<SYS>>
INSTRUCTION
Translate the following clinical note into HL7 FHIR R4 Format. 
- Do not insert any values that are not in the note. 
- Do not infer or impute any values
- Only include information that is essential:
    - information that is in the clinical note
    - information that is mandatory for a valid FHIR resource.

OUTPUT FORMAT
Return the HL7 FHIR structured information as a json string. denote the start and end of the json with a markdown codeblock:
```json 
[RESOURCE HERE]
```
<</SYS>>

CLINICAL NOTE
{note}

[/INST]
~
"""

mixtral_template = """[INST]
INSTRUCTION
Translate the following clinical note into HL7 FHIR R4 Format. 
- Do not insert any values that are not in the note. 
- Do not infer or impute any values
- Only include information that is essential:
    - information that is in the clinical note
    - information that is mandatory for a valid FHIR resource.

OUTPUT FORMAT
Return the HL7 FHIR structured information as a json string. denote the start and end of the json with a markdown codeblock:
```json 
[RESOURCE HERE]
```

CLINICAL NOTE
{note}

[/INST]
~
"""

resource_to_note_template = """
Goal: Convert the following JSON file describing a HL7 FHIR R4 Resource to a clinical note.

Resource: 
{resource}

Details:
- Don't mention ID's, coding concepts or other FHIR related artifacts.
- Explicitely describe all of the information in the resource, including the relationship between resources.
- Datetime fields that are converted to text should still include timezone information and should include seconds.
- Don't use placeholders/imputation
- Don't ask follow-up questions, just convert the resource to human-readible text.
- Information that has the value 'unknown' in the resource don't have to be converted to text.
- {custom_detail_1}
- {custom_detail_2}
- {custom_detail_3}
"""

template_dict = {
    "llama": llama_template,
    "mixtral": mixtral_template,
    "gpt": gpt_template
}


#  Copyright (c) 2024. HealthSage AI.
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Contains all in-scope FHIR models for the note_to_fhir project
"""
from fhir.resources.R4B.reference import Reference
from fhir.resources.R4B.patient import (
    Patient,
    PatientCommunication,
    PatientContact,
    PatientLink,
)
from fhir.resources.R4B.address import Address
from fhir.resources.R4B.codeableconcept import CodeableConcept
from fhir.resources.R4B.coding import Coding
from fhir.resources.R4B.humanname import HumanName
from fhir.resources.R4B.contactpoint import ContactPoint
from fhir.resources.R4B.encounter import Encounter, EncounterParticipant, EncounterHospitalization, EncounterLocation, EncounterDiagnosis
from fhir.resources.R4B.allergyintolerance import AllergyIntolerance, AllergyIntoleranceReaction
from fhir.resources.R4B.period import Period
from fhir.resources.R4B.narrative import Narrative
from fhir.resources.R4B.identifier import Identifier
from fhir.resources.R4B.bundle import Bundle, BundleEntry
from fhir.resources.R4B.organization import Organization, OrganizationContact
from fhir.resources.R4B.practitioner import Practitioner
from fhir.resources.R4B.procedure import Procedure, ProcedurePerformer
from fhir.resources.R4B.condition import Condition
from fhir.resources.R4B.immunization import Immunization, ImmunizationProtocolApplied
from fhir.resources.R4B.observation import Observation, ObservationComponent
from fhir.resources.R4B.medication import Medication
from fhir.resources.R4B.quantity import Quantity
from fhir.resources.R4B.resource import Resource
from fhir.resources.R4B.extension import Extension

object_mapping = {
    "Patient": Patient,
    "PatientCommunication": PatientCommunication,
    "PatientContact": PatientContact,
    "PatientLink": PatientLink,
    "Address": Address,
    "CodeableConcept": CodeableConcept,
    "Coding": Coding,
    "HumanName": HumanName,
    "ContactPoint": ContactPoint,
    "Encounter": Encounter,
    "EncounterParticipant": EncounterParticipant,
    "AllergyIntolerance": AllergyIntolerance,
    "Period": Period,
    "Narrative": Narrative,
    "Identifier": Identifier,
    "Bundle": Bundle,
    "BundleEntry": BundleEntry,
    "Organization": Organization,
    "Practitioner": Practitioner,
    "Procedure": Procedure,
    "Condition": Condition,
    "Immunization": Immunization,
    "Observation": Observation,
    "Medication": Medication,
    "Reference": Reference,
    "Quantity": Quantity,
    "Resource": Resource,
    "AllergyIntoleranceReaction": AllergyIntoleranceReaction,
    "ProcedurePerformer": ProcedurePerformer,
    "ImmunizationProtocolApplied": ImmunizationProtocolApplied,
    "ObservationComponent": ObservationComponent,
    "EncounterHospitalization": EncounterHospitalization,
    "EncounterLocation": EncounterLocation,
    "OrganizationContact": OrganizationContact,
    "EncounterDiagnosis": EncounterDiagnosis,
    "Extension": Extension,
}



import plotly.graph_objects as go
from collections import defaultdict
from healthsageai.note_to_fhir.evaluation.datamodels import FhirDiff
import pprint


def dict_to_html(fhir_dict: dict) -> str:
    """Prettify a FHIR dict for HTML output. Dicts are shortened to 100 characters

    Args:
        fhir_dict (dict): dictionary with FHIR data

    Returns:
        str: HTML string with FHIR data, where newlines and tabs are replaced by <br> and spaces.
    """
    if isinstance(fhir_dict, defaultdict):
        fhir_dict = dict(fhir_dict)
    printer = pprint.PrettyPrinter(depth=1)
    dict_pretty = printer.pformat(fhir_dict)
    dict_pretty = dict_pretty.replace("\t", "  ").replace("\n", "<br>")
    if len(dict_pretty) > 100 and len(dict_pretty.split("<br>")) > 3:
        dict_lines = dict_pretty.split("<br>")
        dict_pretty = (
            dict_lines[0]
            + "<br>"
            + f"<i>...{len(dict_lines)-2} lines...</i>"
            + "<br>"
            + dict_lines[-1]
        )
    return dict_pretty


def preprocess_for_treemap(diff: FhirDiff):
    """Flattens a nested FhirDiff object for plotly treemaps

    Args:
        comparison (FhirDiff): FhirDiff object

    Returns:
        (labels, parents, values):
            - labels define the "name" of the node
            - parents define the parent label
            - values or the content of the treemap
    """

    # Current Node
    labels = [diff.label]
    parents = [diff.parent.label] if diff.parent else [""]
    values = [diff]

    if not diff.children:
        return labels, parents, values

    for child_comparison in diff.children.values():
        if isinstance(child_comparison, list):
            for child_comparison_item in child_comparison:
                new_labels, new_parents, new_values = preprocess_for_treemap(
                    child_comparison_item
                )
                labels = labels + new_labels
                parents = parents + new_parents
                values = values + new_values

        else:
            new_labels, new_parents, new_values = preprocess_for_treemap(
                child_comparison
            )
            labels = labels + new_labels
            parents = parents + new_parents
            values = values + new_values

    return labels, parents, values


def show_diff(diff: FhirDiff) -> None:
    """Produces a treemap visualization of the FhirDiff object in plotly

    Args:
        diff (FhirDiff): Processed FhirDiff object with scores calculated.
    """
    labels, parents, values = preprocess_for_treemap(diff)
    max_level = max([len(x.label.split(".")) for x in values])

    # Shorten text labels for treemap, only use last part of label, or last 2 parts if last part refers to index
    labels_short = [
        label.split(".")[-1]
        if not label.split(".")[-1].isnumeric()
        else label.split(".")[-2:]
        for label in labels
    ]

    fig = go.Figure(
        go.Treemap(
            labels=labels,
            parents=parents,
            values=[max_level + 1 - len(x.label.split(".")) for x in values],
            textinfo="text",
            text=labels_short,
            hovertext=[
                f"{x.score}<br><i>true: </i>{dict_to_html(x.fhir_true)}<br><i>pred: </i>{dict_to_html(x.fhir_pred)}"
                for x in values
            ],
            hoverinfo="text",
            marker_colors=[x.score.accuracy for x in values],
            marker_colorscale="RdBu",
            marker_cmid=0.5,
        )
    )

    fig.update_layout(margin=dict(t=50, l=25, r=25, b=25))
    fig.update_traces(marker=dict(cornerradius=5))
    fig.show()



#  Copyright (c) 2024. HealthSage AI.
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from pydantic import BaseModel, computed_field, field_validator, Field
from typing import Any, Optional
from collections import defaultdict

class FhirValiditionScore(BaseModel):
    n_nodes: int = 0
    n_valid_nodes: int = 0
    n_invalid_nodes: int = 0
    self_is_valid: Optional[bool] = None

    @computed_field
    @property
    def validation_score(self) -> float:
        if self.n_nodes == 0:
            return None
        else:
            return self.n_valid_nodes / self.n_nodes


class FhirScore(BaseModel):
    n_leaves: int = 0  # The number of leaf nodes in this object
    n_additions: int = 0  # n hallucinated leaf nodes, a.k.a. "False Positives"
    n_deletions: int = 0  #  n of missing leaf nodes, a.k.a. "False Negatives"
    n_modifications: int = 0  # n of changes leaf nodes a.k.a. "Mistakes"
    n_matches: int = 0  # n of identical leaf nodes, a.k.a. "True Positives"

    @computed_field
    @property
    def accuracy(self) -> float:  # The overall accuracy
        if self.n_leaves == 0:
            return None
        return self.n_matches / self.n_leaves

    @computed_field
    @property
    def precision(self) -> float:  # What % of generated FHIR was correct
        total = self.n_matches + self.n_additions + self.n_modifications
        if total == 0:
            return None
        return self.n_matches / total

    @computed_field
    @property
    def recall(self) -> float:  # What % of reference FHIR was correctly generated
        total = self.n_matches + self.n_deletions + self.n_modifications
        if total == 0:
            return None
        return self.n_matches / total

    def __add__(self, other: "FhirScore"):
        if not isinstance(other, FhirScore):
            return self
        return FhirScore(
            n_leaves=self.n_leaves + other.n_leaves,
            n_additions=self.n_additions + other.n_additions,
            n_deletions=self.n_deletions + other.n_deletions,
            n_modifications=self.n_modifications + other.n_modifications,
            n_matches=self.n_matches + other.n_matches,
        )

    def __radd__(self, other: Any):
        if not isinstance(other, FhirScore):
            return self
        else:
            return self.__add__(other)


class ElementDetails(BaseModel):
    key: str
    fhirtype: str
    required: bool
    is_leaf: bool
    is_array: bool
    is_struct: bool
    array_item_type: Optional[str]


class FhirDiff(BaseModel):
    fhir_true: Any = Field(repr=False)
    fhir_pred: Any = Field(repr=False)
    resource_name: str  # resource type or fhir type
    parent: Optional["FhirDiff"] = Field(default_factory=lambda: None, repr=False)
    children: dict = Field(default_factory=dict, repr=False)
    entry_nr: str = ""  # For lists, tracking index
    key: str = ""  # What the element is named in its parent object
    score: FhirScore = FhirScore()

    @computed_field
    @property
    def label(self) -> str:
        """Labels are used to identify the node in the FHIR tree

        Returns:
            str: node label describing PARENT.KEY.ENTRY_NR where parent, key and entry_nr are optional
        """
        parent_label = "" if not self.parent else self.parent.label
        keylabel = self.key if self.key != "resource" else self.resource_type
        label = ".".join([parent_label, keylabel, self.entry_nr]).strip(".")
        return label

    @computed_field
    @property
    def resource_type(self) -> str:
        if isinstance(self.fhir_true, dict) and len(self.fhir_true) > 0:
            if self.fhir_true["resourceType"]:
                return self.fhir_true["resourceType"]
        return self.resource_name

    @field_validator("fhir_true", "fhir_pred", mode="before")
    def convert_to_defaultdict(cls, v):
        if isinstance(v, dict):
            return defaultdict(lambda: {}, v)
        if v is None:
            return defaultdict(lambda: {})
        return v
    


#  Copyright (c) 2024. HealthSage AI.
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import json
from healthsageai.note_to_fhir.evaluation.datamodels import FhirScore, ElementDetails, FhirDiff
from healthsageai.note_to_fhir.evaluation.fhirmodels import object_mapping
from typing import List
import warnings
from collections import defaultdict
from pydantic.v1.main import ModelMetaclass
import pandas as pd
import itertools
import numpy as np

MAX_PERMUTATIONS_ARRAY_SIZE = 5  # When all permutations have to be calculated


def get_resource_details(Resource) -> List[ElementDetails]:
    """Get the details of a certain fhir resource that are relevant to evaluation in a friendly format.

    Args:
        Resource (fhir.resources): Resource Metadata

    Returns:
        ResourceDetails: details about a fhir resource
    """

    resource_details = []

    for name, spec in Resource.schema()["properties"].items():
        if "element_property" not in spec.keys():
            continue
        if not spec["element_property"]:
            continue
        if "type" not in spec.keys():
            continue

        required = (
            False if "element_required" not in spec.keys() else spec["element_required"]
        )

        fhirtype = spec["type"]
        array_item_type = None
        if fhirtype == "array":
            array_item_type = spec["items"]["type"]

        element_details = ElementDetails(
            key=name,
            fhirtype=fhirtype,
            required=required,
            is_leaf=fhirtype_is_leaf(fhirtype),
            is_struct=fhirtype_is_struct(fhirtype),
            is_array=fhirtype_is_array(fhirtype),
            array_item_type=array_item_type,
        )

        resource_details.append(element_details)

    return resource_details


def match_list_len(list_1: list, list_2: list) -> tuple:
    """Iteratively appends None to the shortest list untill it matches the longest list
    Args:
        list_1 (list): A list of items
        list_2 (list): A list of items

    Returns:
        tuple: The respective lists but of equal length, imputed with None.
    """
    list_1 = list_1 if list_1 else []
    list_2 = list_2 if list_2 else []
    len_1 = len(list_1)
    len_2 = len(list_2)
    max_len = max(len_1, len_2)
    for i in range(max_len):
        if i >= len_1:
            list_1.append(None)
        elif i >= len_2:
            list_2.append(None)
        else:
            continue
    return list_1, list_2


def validate_resource(resource: dict) -> bool:
    """Checks whether a FHIR resource is valid

    Args:
        resource (dict): A dictionary representing a FHIR R4B resource.

    Returns:
        bool: True if the resource was successfully parsed, False otherwise.
    """
    assert "resourceType" in resource.keys(), "resourceType unspecified"
    try:
        ResourceClass = object_mapping[resource["resourceType"]]
        resource = ResourceClass.parse_raw(json.dumps(resource))
        is_parsed = True
    except Exception:
        is_parsed = False
    return is_parsed


def fhirtype_is_struct(fhirtype: str) -> bool:
    """Check whether fhirtype is a struct by checking if the first letter of the type is a capital letter.

    Args:
        fhirtype (str): fhirtype as specified by fhir.resources

    Returns:
        bool: True if fhirtype is struct, False otherwise
    """
    return fhirtype[0] != fhirtype[0].lower()


def fhirtype_is_array(fhirtype: str) -> bool:
    """Check whether fhirtype is array

    Args:
        fhirtype (str): fhirtype as specified by fhir.resources

    Returns:
        bool: True if fhirtype is array, False otherwise
    """
    return fhirtype == "array"


def fhirtype_is_leaf(fhirtype) -> bool:
    """Check whether a fhirtype is a leaf. i.e. whether the fhirtype contains no nested fields.

    Args:
        fhirtype (str): fhirtype as specified by fhir.resources

    Returns:
        bool: True if fhirtype is leaf, False otherwise
    """
    return fhirtype in ["boolean", "integer", "string", "decimal", "number"]


def map_align_arrays(arr1, arr2):
    """Maps two arrays to each other and aligns them in corresponding order.

    Args:
        arr1 (list): list of fhir elements
        arr2 (list): list of fhir elements

    Returns:
        arr1, arr2: Where arr1[i] corresponds to arr2[i]
    """
    warnings.warn("NotImplemented; arrays are assumed to be in identical order.")
    return match_list_len(arr1, arr2)

def optimize_array_order(fhir_true_array: list, fhir_pred_array: list, element_details: ElementDetails) -> tuple:
    if len(fhir_true_array) <= MAX_PERMUTATIONS_ARRAY_SIZE:
        return optimize_array_order_exact(fhir_true_array, fhir_pred_array, element_details)  # scales n!
    else:
        return optimize_array_order_approx(fhir_true_array, fhir_pred_array, element_details)  # scales n**2

def optimize_array_order_exact(fhir_true_array: list, fhir_pred_array: list, element_details: ElementDetails) -> tuple:
    """Finds the list order for fhir_pred the results in the highest accuracy by calculating all permutations

    Args:
        fhir_true_array (list): list of Fhir resources
        fhir_pred_array (list): list of Fhir resources to optimize
        element_details (ElementDetails): metadata
    """
    assert isinstance(fhir_true_array, list) and isinstance(fhir_pred_array, list), (fhir_true_array, fhir_pred_array)
    fhir_pred_child_permutations = [list(permutation) for permutation in itertools.permutations(fhir_pred_array)]
    max_idx = -1
    max_accuracy = 0.
    for i, permutation in enumerate(fhir_pred_child_permutations):
        score = _get_array_score(fhir_true_array, permutation, element_details)
        if score.accuracy > max_accuracy:
            max_idx = i
            max_accuracy = score.accuracy
    fhir_pred_child_max = fhir_pred_child_permutations[max_idx]
    return fhir_true_array, fhir_pred_child_max

def optimize_array_order_approx(fhir_true_array: list, fhir_pred_array: list, element_details: ElementDetails) -> tuple:
    """Finds the list order for fhir_pred the results in the highest accuracy using argmax with penalties

    Args:
        fhir_true_array (list): _description_
        fhir_pred_array (list): _description_
        element_details (ElementDetails): _description_

    Returns:
        tuple: _description_
    """
    accuracy_matrix = pd.DataFrame(0., index=np.arange(len(fhir_true_array)), columns=np.arange(len(fhir_pred_array)))
    for i_true, fhir_true in enumerate(fhir_true_array):
        for i_pred, fhir_pred in enumerate(fhir_pred_array):
            diff = FhirDiff(
            fhir_true=fhir_true,
            fhir_pred=fhir_pred,
            resource_name=element_details.array_item_type,
            parent=None,
            key=element_details.key,
            )
            diff = _expand_diff_tree(diff)
            accuracy_matrix.iloc[i_true, i_pred] = diff.score.accuracy

    optimal_order = get_optimal_order(accuracy_matrix)
    fhir_pred_array_max = []
    for target_idx in optimal_order:
        fhir_pred_array_max.append(fhir_pred_array[target_idx])

    return fhir_true_array, fhir_pred_array_max


def get_optimal_order(accuracy_matrix):
    """

    Args:
        matrix (_type_): _description_
    """
    epsilon = 0.01
    marginal_distribution_pred = accuracy_matrix.mean(axis=0) + epsilon
    marginal_distribution_true = accuracy_matrix.mean(axis=1) + epsilon
    joint_distribution = np.dot(np.expand_dims(marginal_distribution_true, axis=1), np.expand_dims(marginal_distribution_pred, axis=0))

    accuracy_ratio_matrix = accuracy_matrix / joint_distribution

    optimal_order = accuracy_ratio_matrix.idxmax(axis=1)

    return optimal_order


def _get_array_score(fhir_true_array: list, fhir_pred_array: list, element_details: ElementDetails) -> FhirScore:
    """Calculate FhirScore of fhir_pred_array in that particular order

    Args:
        fhir_true_array (list): list of Fhir resources
        fhir_pred_array (list): list of Fhir resources to optimize
        element_details (ElementDetails): metadata

    Returns:
        FhirScore: FhirScore of fhir_pred_array in that particular order
    """
    i = 0
    childscore = FhirScore()
    for fhir_true_child_item, fhir_pred_child_item in zip(
        fhir_true_array, fhir_pred_array
    ):
        childdiff_item = FhirDiff(
            fhir_true=fhir_true_child_item,
            fhir_pred=fhir_pred_child_item,
            resource_name=element_details.array_item_type,
            parent=None,
            entry_nr=str(i),
            key=element_details.key,
        )
        childdiff_item = _expand_diff_tree(childdiff_item)
        childscore = childscore + childdiff_item.score
        i += 1
    return childscore



def convert_to_defaultdict(obj) -> any:
    """Convert dict to default dict, if it's a dict

    Args:
        obj (any): _description_

    Returns:
        any: defaultdict if the input was a dict, else just returns the input
    """
    if isinstance(obj, dict) or obj is None:
        return defaultdict(dict, obj) if obj else defaultdict(dict)
    else:
        return obj


def get_resource_type(resource, resource_name) -> str:
    """Determine the resource type of a resource

    Args:
        resource (Any): The FHIR resource, can be any datatype.

    Returns:
        str: string representation of the resource type
    """
    if isinstance(resource, dict):
        if resource["resourceType"]:
            return resource["resourceType"]
    return resource_name


def get_resource_class(resource_type: str) -> ModelMetaclass:
    """Get the FhirResource class for a given resource type.

    Args:
        resource_type (str): The resource type

    Returns:
        FhirResource: The FhirResource class
    """
    return object_mapping[resource_type]


def get_diff(fhir_true: dict, fhir_pred: dict, resource_type: str) -> FhirDiff:
    """Calculate the FhirDiff object for comparing two FHIR resources.

    Args:
        fhir_true (dict): The ground truth FHIR resource
        fhir_pred (dict): The predicted/generated FHIR resource
        resource_type (str): The resource type

    Returns:
        FhirDiff: Tree object containing the fhir to be compared.
    """
    diff = FhirDiff(
        fhir_true=fhir_true,
        fhir_pred=fhir_pred,
        resource_name=resource_type,
        key=resource_type,
    )
    diff = _expand_diff_tree(diff)
    return diff


def _expand_diff_tree(diff: FhirDiff) -> FhirDiff:
    """Process FhirDiff to calculate FhirDiff.fhirscore

    Args:
        diff (FhirDiff): comparison object containing the fhir to be compared

    Returns:
        FhirComparison: comparison with fhirscore attribute calculated.
    """
    if diff.fhir_true:
        resource_type = get_resource_type(diff.fhir_true, diff.resource_name)
    else:
        resource_type = get_resource_type(diff.fhir_pred, diff.resource_name)

    if fhirtype_is_leaf(resource_type):
        diff.score = compare_leaf(diff)
        return diff

    Resource = get_resource_class(
        resource_type
    )
    resource_details = get_resource_details(Resource)  # list of ElementDetails



    for element_details in resource_details:
        # Skip if the element is absent in both fhir_true and fhir_pred.
        if (
            element_details.key not in diff.fhir_pred.keys()
            and element_details.key not in diff.fhir_true.keys()
        ):
            continue
        if (
            element_is_absent(diff.fhir_true[element_details.key]) and element_is_absent(diff.fhir_pred[element_details.key])
        ):
            continue

        # If the element is a struct (dictionary) with arbitrary depth, handle recursively
        if element_details.is_struct:
            _expand_diff_tree_struct(diff, element_details)
            childscore = diff.children[element_details.key].score

        # If the element is an array, handle recursively on each array item
        elif element_details.is_array:
            _expand_diff_tree_array(diff, element_details)
            childscore = sum(
                [item.score for item in diff.children[element_details.key]]
            )

        # If the element is a leaf, calculate the score directly
        elif element_details.is_leaf:
            _expand_diff_tree_leaf(diff, element_details)
            childscore = diff.children[element_details.key].score

        else:
            childscore = FhirScore()
            warnings.warn(f"Details of element {element_details} could not be determined. \n fhir true: {diff.fhir_true} \n fhir pred: {diff.fhir_pred}")

        # Add the child node score to the current score
        diff.score = diff.score + childscore

    # diff.score.is_valid = validate_resource(diff.fhir_pred, diff.resource_type)

    return diff


def _expand_diff_tree_leaf(diff: FhirDiff, element_details: ElementDetails):
    """Expands FhirDiff with leaf node

    Args:
        diff (FhirDiff): comparison object containing the fhir to be compared
        element_details (ElementDetails): details about the element to be compared
    """
    childdiff = FhirDiff(
        fhir_true=diff.fhir_true[element_details.key],
        fhir_pred=diff.fhir_pred[element_details.key],
        resource_name=element_details.fhirtype,
        parent=diff,
        key=element_details.key,
    )
    childscore = compare_leaf(childdiff)
    childdiff.score = childscore
    diff.children[element_details.key] = childdiff


def _expand_diff_tree_struct(diff: FhirDiff, element_details: ElementDetails):
    """Expand FhirDiff with struct/dict-like node

    Args:
        diff (FhirDiff): _description_
        element_details (ElementDetails): _description_
    """
    if not isinstance(diff.fhir_pred[element_details.key], dict):
        diff.fhir_pred[element_details.key]= {}
    childdiff = FhirDiff(
        fhir_true=diff.fhir_true[element_details.key],
        fhir_pred=diff.fhir_pred[element_details.key],
        resource_name=element_details.fhirtype,
        parent=diff,
        key=element_details.key,
    )
    childdiff = _expand_diff_tree(childdiff)
    diff.children[element_details.key] = childdiff


def _expand_diff_tree_array(diff: FhirDiff, element_details: ElementDetails):
    """Expand FhirDiff with array node

    Args:
        diff (FhirDiff): _description_
        element_details (ElementDetails): _description_
    """
    if not isinstance(diff.fhir_pred[element_details.key], list):
        diff.fhir_pred[element_details.key] = []
    diff.children[element_details.key] = []
    fhir_true_child, fhir_pred_child = diff.fhir_true[element_details.key], diff.fhir_pred[element_details.key]
    fhir_true_child, fhir_pred_child = match_list_len(fhir_true_child, fhir_pred_child)
    if len(fhir_true_child) > 1 or len(fhir_pred_child) > 1:
        fhir_true_child, fhir_pred_child = optimize_array_order(fhir_true_child, fhir_pred_child, element_details)

    i = 0
    childscore = FhirScore()
    for fhir_true_child_item, fhir_pred_child_item in zip(
        fhir_true_child, fhir_pred_child
    ):
        childdiff_item = FhirDiff(
            fhir_true=fhir_true_child_item,
            fhir_pred=fhir_pred_child_item,
            resource_name=element_details.array_item_type,
            parent=diff,
            entry_nr=str(i),
            key=element_details.key,
        )
        childdiff_item = _expand_diff_tree(childdiff_item)
        diff.children[element_details.key].append(childdiff_item)
        childscore = childscore + childdiff_item.score
        i += 1

def remove_id_from_reference(reference: str):
    """Remove id from reference for evaluation. 
    
    Args:
        reference (str): Reference to another FHIR resource, e.g. "Patient/1", "Encounter/2" etc.

    Returns:
        str: Reference without the id, e.g. "Patient", "Encounter"
    """
    if "/" in reference:
        return reference.split("/")[0]
    else:
        return reference
    
def element_is_absent(leaf: any) -> bool:
    """Check if an element is semantically null/None, taking into account different types.

    Args:
        leaf (any): The value of a Fhir Element

    Returns:
        bool: True if the leaf is empty, False otherwise
    """
    if leaf is None:
        return True
    if leaf == "":
        return True
    if leaf == []:
        return True
    if leaf == {}:
        return True
    else:
        return False


def compare_leaf(diff: FhirDiff) -> FhirScore:
    """Compares two leaf nodes of a FHIR structure

    Args:
        element_true (any): The ground truth fhir element
        element_pred (any): The predicted fhir element

    Returns:
        FhirScore: object containing score for the leaf node.
    """
    element_true, element_pred = diff.fhir_true, diff.fhir_pred
    if diff.key == "id":
        return FhirScore()
    if diff.key == "reference":
        element_true, element_pred = remove_id_from_reference(element_true), remove_id_from_reference(element_pred)
    if element_is_absent(element_pred) and element_is_absent(element_true):
        fhirscore = FhirScore()
    elif element_is_absent(element_pred):
        fhirscore = FhirScore(n_deletions=1, n_leaves=1)  # miss
    elif element_is_absent(element_true):
        fhirscore = FhirScore(n_additions=1, n_leaves=1)  # hallucination
    elif element_true != element_pred:
        fhirscore = FhirScore(n_modifications=1, n_leaves=1)  # mistake
    elif element_true == element_pred:
        fhirscore = FhirScore(n_matches=1, n_leaves=1)  # correct
    return fhirscore


def diff_to_list(diff: FhirDiff) -> list:
    """Flattens a Diff Tree object to a list of unhierarchical FhirDiff objects.

    Args:
        comparison (FhirDiff): FhirDiff object

    Returns:
        list: list of FhirDiff objects
    """

    # Current Node
    diff_copy = diff.model_copy(deep=True)
    diff_copy.parent = None
    diff_copy.children = None
    diffs = [diff_copy]

    if not diff.children:
        return diffs

    for child_comparison in diff.children.values():
        if isinstance(child_comparison, list):
            for child_comparison_item in child_comparison:
                new_diffs = diff_to_list(
                    child_comparison_item
                )
                diffs = diffs + new_diffs

        else:
            new_diffs = diff_to_list(
                child_comparison
            )
            diffs = diffs + new_diffs
    return diffs


def diff_to_dataframe(diff: FhirDiff) -> pd.DataFrame:
    """Flattens a Diff Tree object to a pandas dataframe

    Args:
        comparison (FhirDiff): FhirDiff object

    Returns:
        pd.DataFrame: pandas dataframe containing the diff
    """
    diffs = diff_to_list(diff)
    diff_dicts = []
    scores = [diff.score for diff in diffs]
    for diff in diffs:
        diff_dict_out = diff.model_dump()
        score = diff.score.model_dump()
        diff_dict_out.update(score)
        diff_dicts.append(diff_dict_out)
    df = pd.DataFrame(diff_dicts)[['resource_type', 'entry_nr', 'key', 'label', 'n_leaves', 'n_matches','n_additions','n_deletions','n_modifications','accuracy','precision','recall']]
    df['score'] = scores
    return df



#  Copyright (c) 2023. HealthSage AI.
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import json
import re
import marko


def extract_from_code_block(markdown):
    parsed = marko.Parser().parse(markdown)
    code_block = [elem for elem in parsed.children if isinstance(
        elem, marko.block.FencedCode)]
    if len(code_block) > 0:
        raw_data = code_block[0].children[0].children
        return ''.join(raw_data)
    else:
        return markdown if '\n\n' not in markdown else markdown.split('\n\n')[1]


def parse_json_markdown(markdown_string: str) -> dict:
    json_str = extract_from_code_block(markdown_string)

    return json.loads(json_str)

def parse_json(s) -> str:
    
    # regular expression pattern to find text between ```json and ```
    pattern = re.compile(r'```json(.*?)```', re.DOTALL)
    
    # search the string for the pattern
    match = pattern.search(s)
    
    # check if a match was found
    if match:
        # extract the matched substring
        substring = match.group(1)
    
        # output the extracted substring
        return substring
    else:
        return "{}"
    
def parse_note_to_fhir(s) -> dict:
    """Parse FHIR JSON string from Note-to-Fhir output

    Args:
        s (_type_): _description_

    Returns:
        dict: _description_
    """
    fhir_json = s.split("```")[3][4:].strip(" \t\n\r")
    fhir = json.loads(fhir_json)
    return fhir



#  Copyright (c) 2023. HealthSage AI.
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from transformers import BitsAndBytesConfig, AutoModelForCausalLM, AutoTokenizer, pipeline
import torch
from healthsageai.note_to_fhir.data_utils import drop_snomed_loinc, drop_nones
from healthsageai.note_to_fhir.templates.simple import template_dict
from healthsageai.note_to_fhir.parsers import parse_note_to_fhir

class NoteToFhir(object):

    def __init__(self, model_name: str, adapter_name: str, template_style: str) -> None:
        """_summary_

        Args:
            model_name (str or os.PathLike): The base model
            adapter_name (str or os.PathLike): The Q-LoRA adapter
            template_style (str): "gpt", "llama" or "mixtral"
        """
        self.template = template_dict[template_style]
        bnb_config = BitsAndBytesConfig(
        load_in_4bit=True,
        bnb_4bit_use_double_quant=True,
        bnb_4bit_quant_type="nf4",
        bnb_4bit_compute_dtype=torch.float16,
        )

        model = AutoModelForCausalLM.from_pretrained(
            model_name,
            trust_remote_code=True,
            quantization_config=bnb_config,
            device_map='auto',
        )

        model.config.use_cache = False
        model.load_adapter(adapter_name)

        tokenizer = AutoTokenizer.from_pretrained(model_name, trust_remote_code=True, return_tensor="pt", padding=True)
        tokenizer.pad_token = tokenizer.bos_token
        tokenizer.padding_side = 'right'

        self.generator = pipeline(
            model=model, 
            tokenizer=tokenizer,
            task="text-generation",
            do_sample=True,
            eos_token_id=model.config.eos_token_id,
            max_length=4096
        )

    def translate(self, note : str) -> dict:
        """Convert a note to FHIR

        Args:
            note (str): clinical note
        """
        prompt = self.template.format(note=note)
        generated_output = self.generator(prompt)
        fhir = parse_note_to_fhir(generated_output[0]['generated_text'])
        fhir = drop_nones(fhir)
        fhir = drop_snomed_loinc(fhir)
        return fhir
    
class NoteToFhir13b(NoteToFhir):

    def __init__(self):
        super().__init__(model_name="meta-llama/Llama-2-13b-chat-hf",
                        adapter_name="healthsageai/note-to-fhir-13b-adapter",
                        template_style="llama")
        
class NoteToFhir8x7b(NoteToFhir):

    def __init__(self):
        super().__init__(model_name="mistralai/Mixtral-8x7B-Instruct-v0.1",
                        adapter_name="healthsageai/note-to-fhir-8x7b-mixtral-dev",
                        template_style="mixtral")


#  Copyright (c) 2023. HealthSage AI.
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

def drop_code(coding: dict) -> dict:
    """Drop code for SNOMED and Loinc codings

    Args:
        coding (dict): coding

    Returns:
        dict: Copy of coding with SNOMED and Loinc codes deleted
    """
    coding_out = coding.copy()
    if "code" in coding.keys():
        del coding_out['code']
    return coding_out

def drop_snomed_loinc(d):
    """Recursively drop Nones in dict d and return a new dict"""
    dd = {}
    for k, v in d.items():
        if isinstance(v, dict):
            dd[k] = drop_snomed_loinc(v)
        elif isinstance(v, list) and k != "coding":
            # note: Nones in lists are not dropped
            dd[k] = [drop_snomed_loinc(vv) if isinstance(vv, dict) else vv
                            for vv in v]
        elif k == 'coding' and isinstance(v, list):
            dd[k] = [drop_code(code) if "snomed" in code.get("system","") or "loinc" in code.get("system","") else code for code in v]
        else:
            dd[k] = v
            
    return dd


def remove_snomed_loinc_code_for_coding(key, val):
    if key == 'code' and isinstance(val['coding'], list):
        val['coding'] = [code for code in val['coding'] if
                         code['system'] not in ['http://snomed.info/sct', 'http://loinc.org']]

    return val


def clean_fhir_resource(fhir: dict) -> dict:
    return {key: remove_snomed_loinc_code_for_coding(key, val) for (key, val) in drop_nones(fhir).items()}


def filter_supported_fhir_resources(resources: list) -> list:
    return [res for res in resources if
            res['resourceType'] in ['Observation', 'Condition', 'Procedure', 'Encounter', 'Patient', 'Organization',
                                    'Location', 'Practitioner', 'Immunization', 'AllergyIntolerance']]


def drop_nones(d: dict) -> dict:
    """Recursively drop Nones in dict d and return a new dict"""
    dd = {}
    for k, v in d.items():
        if isinstance(v, dict):
            dd[k] = drop_nones(v)
        elif isinstance(v, (list, set, tuple)):
            # note: Nones in lists are not dropped
            dd[k] = type(v)(drop_nones(vv) if isinstance(vv, dict) else vv
                            for vv in v)
        elif v is not None:
            dd[k] = v
    return dd
